"""
PlaningHullPy
=============

This is a python library for calculating the dynamics of planing hull
"""
from .core_class import *
from .general_func import *
from .orientation import *
from .constant import *

from . import hydro



import numpy as np
import math

__all__ =['deg_to_rad', 'rad_to_deg']

def _mat_rotation(theta_d):
    """Convert angle into rotation matrix. Input is in degree"""
    theta = theta_d*math.pi/180
    return np.array([[np.cos(theta), -np.sin(theta)],[np.sin(theta), np.cos(theta)]])

def deg_to_rad(theta_d):
    return theta_d*math.pi/180

def rad_to_deg(theta):
    return theta*180/math.pi


import numpy as np
import math

from .general_func import *
from .orientation import *

from . import hydro
from . import constant

class Craft(object):
    """
    Creates a craft object 
    
    Parameters
    -------------------------------------
    name -> name of the craft
    LCG -> Longitudinal CG of the craft
    VCG -> Vertical CG of the craft
    
    List of Variables that can be defined
    -------------------------------------
    mass
    i_yy
    """
    
    # The following value is to be initialized later on
    mass = 1000 # kg
    i_yy = 1000 * (0.25 * 10) ** 2 # kg*m^2
    
    def __init__(self, name='craft', lcg=0, vcg=0):
        self.name = name
        self.lcg  = lcg
        self.vcg  = vcg

        """
        The list of hull of the craft.
        hull[0] is reserved for calm water
        hull[1] will be the main hull, referred most of the time by 'main'
        """
        self.hull =[] # A list of hull
        self.hull.append(Hull('calm_water', 0, 0))
        
        # Initialize state and orientation
        self.state       = State()
        self.wetted      = hydro.Wetted()
        self.hydroFM     = hydro.HydroForceMoment()
        self.main_to_cg  = Orientation(0,  self.lcg,  self.vcg)
        self.cg_to_main  = Orientation(0, -self.lcg, -self.vcg)
        self.main_orient = Orientation(0, 0, 0)
        self.cg_orient   = Orientation(0, 0, 0)
        
    # Printing out the relevant information
    def __call__(self):
        print('The craft information:')
        print('Name = ',             self.name)
        print('LCG  = %8.2f m'      %self.lcg)
        print('VCG  = %8.2f m'      %self.vcg)
        print('Mass = %8.2f kg'     %self.mass)
        print('I_yy = %8.2f kg*m^2' % self.i_yy)
        
    # Adding hull
    def add_hull(self, name, beam, beta_d, warp_b = 0, pos_x = 0, pos_z = 0, relative_trim_d = 0, wake_of= 0):
        available_id = len(self.hull)
        
        # Make sure that each hull is a wake of other hull
        if wake_of > available_id:
            return 'The hull %d is not in the wake of a valid hull' %available_id
        else: 
            self.hull.append(Hull(name, beam, beta_d, warp_b, pos_x, pos_z, relative_trim_d, wake_of, hull_id = available_id))
    
    def init_orient(self):
        """Initialize the orientation of the craft and its hull"""
        
        # Need to get the main_orient first
        trim = math.pi/180*self.state.trim_d
        pos_x = self.lcg*np.cos(trim) - self.vcg*np.sin(trim)

        self.cg_orient = Orientation(self.state.trim_d, pos_x, self.state.z)
        self.main_orient = orient_mul(self.cg_orient, self.cg_to_main)

        no_of_hull = len(self.hull) - 1
        for i in range(1, no_of_hull+1): # Remember hull_id starts with 1
            self.hull[i].hull_orient = orient_mul(self.main_orient, self.hull[i].main_to_hull)
    
    def init_hull_state(self):
        """
        Initialize the state of the craft's hull. 
        Needs to be done after initializing the orientation
        """
        
        no_of_hull = len(self.hull) - 1
        for i in range(1, no_of_hull+1): # Remember hull_id starts with 1
            if self.hull[i].wake_of == 0:
                v      = self.state.v
                trim_d = self.hull[i].hull_orient.theta_d
                pos_z  = self.hull[i].hull_orient.pos_z
                self.hull[i].state.set_state([v, trim_d, pos_z, 0, 0])
    
    def get_hydro(self):
        """get hydro from all hull"""
        
        # First is to reset value to 0 first
        self.hydroFM.reset()
        self.init_orient()
        self.init_hull_state()
        
        no_of_hull = len(self.hull) - 1
        for i in range(1, no_of_hull+1): # Remember hull_id starts with 1
            self.hull[i].get_wetted()
            self.hull[i].get_hydro()
            
            self.hydroFM.lift_pres  += self.hull[i].hydroFM.lift_pres
            self.hydroFM.lift_visc  += self.hull[i].hydroFM.lift_visc
            self.hydroFM.lift_total += self.hull[i].hydroFM.lift_total
            self.hydroFM.drag_pres  += self.hull[i].hydroFM.drag_pres
            self.hydroFM.drag_visc  += self.hull[i].hydroFM.drag_visc
            self.hydroFM.drag_total += self.hull[i].hydroFM.drag_total
    
            x_hull = self.hull[i].hull_orient.pos_x
            z_hull = self.hull[i].hull_orient.pos_z
            z_main = self.main_orient.pos_z
            self.hydroFM.moment   += self.hull[i].hydroFM.moment + \
                                   self.hull[i].hydroFM.lift_total * x_hull + self.hull[i].hydroFM.drag_total * (z_hull - z_main)
        
        # This is not really important, but a representation is given
        self.hydroFM.vcp = self.hull[1].hydroFM.vcp
        self.hydroFM.lcp = (self.hydroFM.moment - self.hydroFM.vcp * self.hydroFM.drag_total) / self.hydroFM.lift_total
        
class State(object):
    """
    The hydrodynamic state is given by 
    v             (m/s)
    trim_d        (deg)
    z             (m)
    trim_d_dot    (deg/s)
    z_dot         (m/s)
    """
    
    def __init__(self):
        self.v          = 10   # m/s
        self.trim_d     = 4    # deg
        self.z          = 0    # m
        self.trim_d_dot = 0    # deg/s
        self.z_dot      = 0    # m/s
        
    def __call__(self, array = None):
        """
        If there is no input, state will be printed.
        If there is any input, then the state will be the output in an array format.
        """
        if array is None:
            print('The state is:')
            print('V         = %6.2f m/s'   %self.v)
            print('trim      = %6.2f deg'   %self.trim_d)
            print('z         = %6.2f m'     %self.z)
            print('trim_dot  = %6.2f deg/s' %self.trim_d_dot)
            print('z_dot     = %6.2f m/s'   %self.z_dot)
        else:
            x = np.zeros(5)
            x[0] = self.v
            x[1] = self.trim_d
            x[2] = self.z
            x[3] = self.trim_d_dot
            x[4] = self.z_dot
            return x
        
    def set_state(self, x):
        """
        Convert an array data to the state
        """
        self.v          = x[0]
        self.trim_d     = x[1]
        self.z          = x[2]
        self.trim_d_dot = x[3]
        self.z_dot      = x[4]
        
class Hull(object):
    """    
    Creates an hull object 
    
    Parameters
    -------------------------------------
    name -> name of the hull
    beam -> width of the beam (m)
    beta_d -> deadrise angle (deg) 
    warp_b -> deadrise warping (deg/beam)
    pos_x -> relative position in x-axis (horizontal) positive is in front
    pos_z -> relative position in z-axis (vertical) positive is upwards
    relative_trim_d -> the relative trim of this hull w.r.t the main hull (deg)
    hull_id -> id of the hull, 1 corresponds to the main hull
    wake_of_hull -> the hull right in front of the existing hull
    
    """
    
    def __init__(self, name, beam, beta_d, warp_b = 0, pos_x = 0, pos_z = 0, relative_trim_d = 0, wake_of = 0, hull_id = 0):
        self.name            = name
        self.beam            = beam
        self.beta_d          = beta_d
        self.warp_b          = warp_b
        self.pos_x           = pos_x
        self.pos_z           = pos_z
        self.relative_trim_d = relative_trim_d
        self.id              = hull_id
        self.wake_of         = wake_of
        
        # Initialize state and orientation
        self.state        = State()
        self.wetted       = hydro.Wetted()
        self.hydroFM      = hydro.HydroForceMoment()
        self.main_to_hull = Orientation(relative_trim_d, pos_x, pos_z)
        self.hull_orient  = Orientation(relative_trim_d, pos_x, pos_z)

        # Initialize Hydro Method
        self.method       = hydro.HydroMethod()

    def __call__(self):
        print('The hull information:')
        print('Name            = ',               self.name)
        print('id              = %8.2f '         %self.id)
        print('beam            = %8.2f m'        %self.beam)
        print('beta_d          = %8.2f deg'      %self.beta_d)
        print('warp_b          = %8.2f deg/beam' %self.warp_b)
        print('relative_trim_d = %8.2f deg'      %self.relative_trim_d)
        print('wake_of hull id = %8.0f'          %self.wake_of)
        
    def get_wetted(self):
        """
        To calculate the non-dimensional aspect of the wetted area.
        Input 
        
        NOW THERE IS NO WARP FOR NOW
        """

        _wetted = hydro.wetted.savitsky_wetted(self)

        self.wetted.cv       = _wetted['cv']
        self.wetted.trim_k_d = _wetted['trim_k_d']
        self.wetted.trim_e_d = _wetted['trim_e_d']
        self.wetted.beta_e_d = _wetted['beta_e_d']
        self.wetted.lambda_k = _wetted['lambda_k']
        self.wetted.lambda_m = _wetted['lambda_m']
        self.wetted.lambda_c = _wetted['lambda_c']
        self.wetted.beam_mul = _wetted['beam_mul']
        
    def get_hydro(self):
        """
        To calculate the force and moment on the hull
        """

        _hydroFM = hydro.get_hydroFM(self)

        self.hydroFM.lift_pres  = _hydroFM['lift_pres']
        self.hydroFM.lift_visc  = _hydroFM['lift_visc']
        self.hydroFM.lift_total = _hydroFM['lift_total']
        self.hydroFM.drag_pres  = _hydroFM['drag_pres']
        self.hydroFM.drag_visc  = _hydroFM['drag_visc']
        self.hydroFM.drag_total = _hydroFM['drag_total']
        self.hydroFM.moment     = _hydroFM['moment']
        self.hydroFM.lcp        = _hydroFM['lcp']
        self.hydroFM.vcp        = _hydroFM['vcp']

def craft_eom(state, craft):
    """
    The EOM of a craft with multi hull

    state here is an array, not an object (MAYBE IN THE FUTURE CAN BE BOTH?)
    """

    # ===========
    # The state
    # ===========
    v          = state[0]
    trim_d     = state[1]
    z          = state[2]
    trim_d_dot = state[3]
    z_dot      = state[4]

    # ==============
    # Final values
    # ==============
    # NOT NOW, LATER MAYBE

    # TO CALCULATE THE HYDRO FORCE
    craft.state.set_state(state)
    craft.get_hydro()

    # ===================
    # Equation of Motion
    # ===================
    cg_x = craft.cg_orient.pos_x
    cg_z = craft.cg_orient.pos_z
    force_z   = craft.hydroFM.lift_total - craft.mass * 9.81
    moment_cg = craft.hydroFM.moment - craft.hydroFM.lift_total * cg_x - craft.hydroFM.drag_total * cg_z

    # Artificial Damping to stabilize
    zeta_f = 20 * max(1, abs(craft.hydroFM.lift_total / craft.mass))
    zeta_m = 8 * max(1, abs(moment_cg / craft.i_yy))

    dx0 = 0  # Speed is constant
    dx1 = trim_d_dot
    dx2 = z_dot
    dx3 = -zeta_m * (trim_d_dot * math.pi / 180) + moment_cg / craft.i_yy
    dx4 = -zeta_f * z_dot + force_z / craft.mass

    return np.array((dx0, dx1, dx2, dx3*180/math.pi, dx4))


import numpy as np
import math

from hydroplaningpy import constant
from hydroplaningpy.general_func import *

# Apart from the force and moment class, we still need to have many of the force related calculation here

class HydroForceMoment(object):
    """A class for hydro force and moment"""

    def __init__(self):
        self.lift_pres = 0
        self.lift_visc = 0
        self.lift_total = 0
        self.drag_pres = 0
        self.drag_visc = 0
        self.drag_total = 0
        self.moment = 0
        self.lcp = 0
        self.vcp = 0

    def __call__(self, array=None):
        """
        If there is no input, state will be printed.
        If there is any input, then the state will be the output in an array format.
        """
        if array is None:
            print('The hydro force is:')
            print('Lift by pressure  = %8.2f N' % self.lift_pres)
            print('Lift by viscosity = %8.2f N' % self.lift_visc)
            print('Total Lift        = %8.2f N' % self.lift_total)
            print('Drag by pressure  = %8.2f N' % self.drag_pres)
            print('Drag by viscosity = %8.2f N' % self.drag_visc)
            print('Total Drag        = %8.2f N' % self.drag_total)
            print('Total Moment      = %8.2f Nm'% self.moment)
            print('Longitudinal CP   = %8.2f m' % self.lcp)
            print('Vertical CP       = %8.2f m' % self.vcp)
        else:
            x = np.zeros(9)
            x[0] = self.lift_pres
            x[1] = self.lift_visc
            x[2] = self.lift_total
            x[3] = self.drag_pres
            x[4] = self.drag_visc
            x[5] = self.drag_total
            x[6] = self.moment
            x[7] = self.lcp
            x[8] = self.vcp
            return x

    def reset(self):
        self.lift_pres = 0
        self.lift_visc = 0
        self.lift_total = 0
        self.drag_pres = 0
        self.drag_visc = 0
        self.drag_total = 0
        self.moment = 0
        self.lcp = 0
        self.vcp = 0

class Wetted(object):
    """
    The non_dimensional state of a hull
    CV
    trim_k_d
    trim_e_d
    beta_e_d
    lambda_k
    lambda_m
    lambda_c
    beam_mul
    """

    def __init__(self, cv=0, trim_k_d=0, trim_e_d=0, beta_e_d=0,
                 lambda_k=0, lambda_m=0, lambda_c=0, beam_mul=0):
        self.cv = cv
        self.trim_k_d = trim_k_d  # deg
        self.trim_e_d = trim_e_d  # deg
        self.beta_e_d = beta_e_d  # deg
        self.lambda_k = lambda_k
        self.lambda_m = lambda_m
        self.lambda_c = lambda_c
        self.beam_mul = beam_mul

    def __call__(self, array=None):
        """
        If there is no input, state_nd will be printed.
        If there is any input, then the state will be the output in an array format.
        """
        if array is None:
            print('The state_nd is:')
            print('CV       = %8.2f' % self.cv)
            print('trim_k_d = %8.2f deg' % self.trim_k_d)
            print('trim_e_d = %8.2f deg' % self.trim_e_d)
            print('beta_e_d = %8.2f deg' % self.beta_e_d)
            print('lambda_k = %8.2f ' % self.lambda_k)
            print('lambda_m = %8.2f ' % self.lambda_m)
            print('lambda_c = %8.2f ' % self.lambda_c)
            print('beam_mul = %8.2f ' % self.beam_mul)
        else:
            x = np.zeros(8)
            x[0] = self.cv
            x[1] = self.trim_k_d
            x[2] = self.trim_e_d
            x[3] = self.beta_e_d
            x[4] = self.lambda_k
            x[5] = self.lambda_m
            x[6] = self.lambda_c
            x[7] = self.beam_mul
            return x

class HydroMethod(object):
    """
    List of hydrodynamic method

    HAS TO BE EXPLAINED HERE LATER
    """
    def __init__(self):
        # These are the default values
        self.wetted         = 1 # 1=Savistky 1964 ; 2=Savitsky 1976
        self.dry_chine      = 1 # 1=Savitsky 1964 correction ; 2=Milwitzky 1948 ; 3=Milwitzky CFD corrected
        self.warp_lambda    = 1 # 1 = Existing equivalent method ; 2=Wagner pi/2 wave rise factor
        self.symmetric_hull = 1 # 1=Half of symmetric hull; 2=Morabito 2011 formulation
        self.roughness = 0.0004 # 0.0004 is often used in full scale
        self.wake_profile   = 1 # 1=Flat from step; 2=Savitsky 2010; 3=Korvin ;4=CFD data; 5=TEST
        self.wake_method    = 1 # 1=Savitsky; 2=KHB; 3=Svahn Method; 4=KHB_2; 5=SVAHN_2 ;6=KHB_3
        self.hum_correction = 0 # 0=No correction  %1=Blount 1976 %2=Savitsky 1976(NOT YET IMPLEMENTED)
        self.whisker_spray  = 1 # 0=Not Calculated %1=Savitsky 2007
        self.spray_method   = 0 # 0=Ignored  %1=Savitsky Chine %2=CFD generated step
        self.spray_sf       = 1 # Safety factor for the spray height

    def __call__(self):
        print('This is to be expanded later!')

def get_hydroFM(hull):
    """
    To calculate the force and moment given hull with its wetted condition
    :param hull:
    :return: hydroFM dictionary
    """

    # Important info
    v    = hull.state.v
    beam = hull.beam

    cv = hull.wetted.cv
    trim_e_d = hull.wetted.trim_e_d
    trim_e   = deg_to_rad(trim_e_d)
    beta_e_d = hull.wetted.beta_e_d
    beta_e   = deg_to_rad(beta_e_d)
    lambda_m = hull.wetted.lambda_m

    roughness = hull.method.roughness

    # Force and Moment Calculation

    cl_0 = trim_e_d ** 1.1 * (0.012 * lambda_m ** 0.5 + 0.0055 * lambda_m ** 2.5 / cv ** 2)
    cl_beta = cl_0 - 0.0065 * beta_e_d * cl_0 ** 0.6
    lift_pres = 0.5 * constant.density_water * v ** 2 * beam ** 2 * cl_beta
    drag_pres = lift_pres * np.tan(trim_e)

    # Calculating viscous force
    cl_0_dyn = trim_e_d ** 1.1 * 0.012 * lambda_m ** 0.5
    cl_beta_dyn = cl_0_dyn - 0.0065 * beta_e_d * cl_0_dyn ** 0.6
    v_1 = v * math.sqrt(1 - cl_beta_dyn / lambda_m / np.cos(trim_e))
    reynolds_no = constant.density_water * v_1 * lambda_m * beam / 0.00122
    c_f = 0.075 / (math.log10(reynolds_no) - 2) ** 2 + roughness

    visc_force = 0.5 * 1025 * v_1 ** 2 * (lambda_m * beam ** 2) / np.cos(beta_e) * c_f
    lift_visc = -visc_force * np.sin(trim_e)
    drag_visc = visc_force * np.cos(trim_e)

    lift_total = lift_pres + lift_visc
    drag_total = drag_pres + drag_visc

    lcp = (0.75 - 1 / (5.21 * (cv / lambda_m) ** 2 + 2.39)) * lambda_m * beam
    vcp = 0.25 * np.tan(beta_e) * beam

    # Need to calculate the moment here USE TRIM_E AS REFERENCE
    # The moment is calculated w.r.t the transom of the hull
    moment = lift_total * (np.cos(trim_e) * lcp - np.sin(trim_e) * vcp) + drag_total * (np.sin(trim_e) * lcp - np.cos(trim_e) * vcp)

    hydroFM = {
        'lift_pres'  : lift_pres,
        'lift_visc'  : lift_visc,
        'lift_total' : lift_total,
        'drag_pres'  : drag_pres,
        'drag_visc'  : drag_visc,
        'drag_total' : drag_total,
        'moment'     : moment,
        'lcp'        : lcp,
        'vcp'        : vcp
    }

    return hydroFM


"""
To handle all related to the orientation matrix.
This orientation matrix is basically a homogeneous transformation for 2D case
"""
import numpy as np
from .general_func import *

__all__ =['Orientation', 'orient_mul', 'zmain_to_zcg', 'zcg_to_zmain']

class Orientation(object):
    """
    Orientation here is a 2D homogeneous transformation matrix.
    It is useful to denote the attitude of a hull as well as transformation between them.
    """
    def __init__(self, theta_d, pos_x, pos_z):
        self.theta_d = theta_d
        self.theta   = deg_to_rad(theta_d)
        self.pos_x   = pos_x
        self.pos_z   = pos_z

    def __call__(self, array = None):
        """
        If there is no input then the state will be printed.
        If there is any input, then the state will be the output in an array format.
        """
        if array is None:
            print('The orientation is:')
            print('theta_d    = %6.2f deg' %self.theta_d)
            print('theta      = %6.2f rad' %self.theta)
            print('x position = %6.2f m'   %self.pos_x)
            print('z position = %6.2f m'   %self.pos_z)
        else:
            self._orient = np.identity(3)
            self._orient[0,2] = self.pos_x
            self._orient[1,2] = self.pos_z
            self._orient[:2,:2] = np.array([[np.cos(self.theta), -np.sin(self.theta)],
                                            [np.sin(self.theta), np.cos(self.theta)]])
            return self._orient

def orient_mul(orient_1, orient_2):
    """
    Basically a wrapper for matrix multiplication for orientation object
    orient_1: Orient2D, object
            The first orientation

    orient_2: Orient2D, object
            The second orientation
    :return:
    """
    theta_d = orient_1.theta_d + orient_2.theta_d
    orient_mat = np.matmul(orient_1(0), orient_2(0)) # Returns a 3x3 array
    pos_x = orient_mat[0,2]
    pos_z = orient_mat[1,2]
    return Orientation(theta_d, pos_x, pos_z)

def zcg_to_zmain(craft, z_cg, trim_d):
    """
    A function that transforms z_cg to the z of the main hull
    """
    cg_to_main = Orientation(0, -craft.lcg, -craft.vcg)
    cg_orient  = Orientation(trim_d, 0, z_cg)
    main_orient = orient_mul(cg_orient, cg_to_main)
    return main_orient.pos_z

def zmain_to_zcg(craft, z_main, trim_d):
    """
    A function that transforms z of the main hull to the z_cg
    """
    main_to_cg  = Orientation(0, craft.lcg, craft.vcg)
    main_orient = Orientation(trim_d, 0, z_main)
    cg_orient   = orient_mul(main_orient, main_to_cg)
    return cg_orient.pos_z


import numpy as np
import math
from hydroplaningpy.general_func import *
from hydroplaningpy import constant

# To calculate wetted area with different method

## To extend into warped hull

def savitsky_wetted(hull):
    """
    This is a function based on Daniel Savitsky 1964 / 1976 paper to calculate wetted area
    :param hull: Hull Object
    :return: wetted area information
    """
    v    = hull.state.v
    beta_d = hull.beta_d
    beta   = deg_to_rad(beta_d)
    trim_d = hull.state.trim_d
    trim   = deg_to_rad(trim_d)

    cv = v/math.sqrt(constant.gravity*hull.beam)

    lambda_k     = -hull.state.z / np.sin(trim) / hull.beam

    if hull.method.wetted == 1: # Based on Savitsky 1964
        w        = np.tan(beta)/(math.pi*np.tan(trim))
        lambda_m = lambda_k - 0.5*w
        lambda_c = lambda_k - w

    elif hull.method.wetted == 2: # Based on Savitsky 1976
        w = (0.57 + beta_d/1000)*(math.tan(beta)/2/math.tan(trim) - beta_d/167)
        lambda_c = lambda_k - w

        if (lambda_c <1):
            lambda_c = (lambda_k - w) - 0.2*math.exp(-(lambda_k - w)/0.3)

        lambda_m = (lambda_k + lambda_c)/2 + 0.03
    else:
        print('Method of calculating wetted area does not exist.')
        return

    # Checking whether chines is dry or wet
    if lambda_c < 0:
        lambda_c = 0
        lambda_m = 0.5 * lambda_k
        beam_mul = lambda_k / w
    else:
        beam_mul = 1

    trim_k_d = hull.state.trim_d
    trim_e_d = trim_k_d  # FOR NOW
    beta_e_d = hull.beta_d

    wetted = {
        'cv' : cv,
        'trim_k_d' : trim_k_d,
        'trim_e_d' : trim_e_d,
        'beta_e_d' : beta_e_d,
        'lambda_k' : lambda_k,
        'lambda_m' : lambda_m,
        'lambda_c' : lambda_c,
        'beam_mul' : beam_mul
    }

    return wetted